#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <malloc.h>
#include <stdbool.h>

#define ROM_SIZE 4 * 1024 * 1024

#define VERSION_STRING "0.2"

size_t find_rom_offset(void * buffer, size_t buffer_size)
{
	uint8_t * bufferptr = (uint8_t *) buffer;
	const uint8_t magic[8] = { 0, 0, 0, 0, 'W', 'U', 'P', '-' };

	size_t position = 0;

	while (position < buffer_size)
	{
		if (memcmp(bufferptr + position, magic, 8) == 0)
		{
			// Skip over the magic 16 bytes
			// e.g. 00 00 00 00 W U P - <etc.>
			return position + 16;
		}

		position += 8;
	}

	return -1;
}

int main(int argc, char** argv)
{
	int return_code = 0;
	FILE * elf_file = NULL;
	FILE * rom_file = NULL;
	FILE * out_file = NULL;
	uint8_t * elf_file_buffer = NULL;
	uint8_t * rom_file_buffer = NULL;

	if ((argc < 4) || (argc > 4))
	{
		printf("(S)NESInject %s: An SNES/NES ROM injector for Wii U VC\n", VERSION_STRING);
		printf("usage: %s <in elf file> <snes rom> <out elf file>\n", argv[0]);
		return_code = -1;
		goto cleanup;
	}

	elf_file = fopen(argv[1], "rb");
	if (elf_file == NULL)
	{
		printf("Unable to open ELF file %s for reading\n", argv[1]);
		return_code = -2;
		goto cleanup;
	}

	size_t elf_size = (size_t) -1;
	fseek(elf_file, 0, SEEK_END);
	elf_size = ftell(elf_file);
	fseek(elf_file, 0, SEEK_SET);

	elf_file_buffer = malloc(elf_size);
	if (elf_file_buffer == NULL)
	{
		printf("Unable to allocate memory for ELF file buffer\n");
		return_code = -3;
		goto cleanup;
	}

	if (fread(elf_file_buffer, 1, elf_size, elf_file) != elf_size)
	{
		printf("Error reading ELF file\n");
		return_code = -4;
		goto cleanup;
	}

	fclose(elf_file);

	printf("Opened ELF file for reading\n");

	rom_file = fopen(argv[2], "rb");
	if (rom_file == NULL)
	{
		printf("Unable to open SNES ROM %s for reading\n", argv[2]);
		return_code = -5;
		goto cleanup;
	}

	size_t rom_size = (size_t) -1;
	fseek(rom_file, 0, SEEK_END);
	rom_size = ftell(rom_file);
	fseek(rom_file, 0, SEEK_SET);

	rom_file_buffer = NULL;
	if (rom_size < ROM_SIZE)
	{
		size_t rom_size_diff = (ROM_SIZE - rom_size);
		rom_file_buffer = malloc(rom_size + rom_size_diff);
		printf("ROM is smaller than 4MB, will attempt to expand by 0x%llx bytes\n", rom_size_diff);
	}
	else if(rom_size == ROM_SIZE)
	{
		rom_file_buffer = malloc(rom_size);
	}
	else
	{
		printf("ROM is too large to be injected, sorry\n");
		return_code = -6;
		goto cleanup;
	}

	if (rom_file_buffer == NULL)
	{
		printf("Unable to allocate memory for ELF file buffer\n");
		return_code = -7;
		goto cleanup;
	}

	// The buffer should end up always being ROM_SIZE in size,
	// so we go ahead and fill it with zeroes for "padding".
	// Automatically handles smaller ROMs basically, and I don't care
	// if fread overwrites the first part of the buffer.
	memset(rom_file_buffer, 0, ROM_SIZE);

	if (fread(rom_file_buffer, 1, rom_size, rom_file) != rom_size)
	{
		printf("Error reading SNES ROM file\n");
		return_code = -8;
		goto cleanup;
	}

	fclose(rom_file);
	printf("Opened SNES ROM file for reading\n");

	size_t rom_offset = find_rom_offset(elf_file_buffer, elf_size);
	if (rom_offset == -1)
	{
		printf("Unable to find WUP- magic; we can't continue, sorry\n");
		return_code = -9;
		goto cleanup;
	}


	memcpy(&elf_file_buffer[rom_offset], rom_file_buffer, ROM_SIZE);

	printf("Injecting ROM into ELF\n");

	out_file = fopen(argv[3], "wb");
	if (out_file == NULL)
	{
		printf("Unable to open out ELF file %s for writing\n", argv[3]);
		return_code = -10;
		goto cleanup;
	}

	if (fwrite(elf_file_buffer, 1, elf_size, out_file) != elf_size)
	{
		printf("Error writing out ELF file\n");
		return_code = -11;
		goto cleanup;
	}

	fclose(out_file);
	printf("Done!\n");

cleanup:
	if (rom_file != NULL)
		fclose(rom_file);

	if (elf_file != NULL)
		fclose(elf_file);

	if (out_file != NULL)
		fclose(out_file);

	free(rom_file_buffer);
	free(elf_file_buffer);

	return return_code;
}
